@@ -3,20 +3,21 @@
 from decimal import Decimal
 from itertools import chain
 
-from trytond import backend
-from trytond.i18n import gettext
-from trytond.pool import PoolMeta, Pool
-from trytond.pyson import Eval, Bool, If, Id
-from trytond.model import Workflow, ModelView, fields, ModelSQL, \
-        sequence_ordered
-from trytond.model.exceptions import RequiredValidationError, AccessError
-from trytond.wizard import Wizard
-from trytond.transaction import Transaction
-from trytond.tools import grouped_slice, firstline
+from sql.functions import CharLength
 
+from trytond.i18n import gettext
+from trytond.model import (
+    Index, ModelSQL, ModelView, Workflow, fields, sequence_ordered)
+from trytond.model.exceptions import AccessError, RequiredValidationError
 from trytond.modules.company.model import (
-    employee_field, set_employee, reset_employee)
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits, round_price
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, Id, If
+from trytond.tools import firstline, grouped_slice
+from trytond.transaction import Transaction
+from trytond.wizard import Wizard
 
 
 class Configuration(metaclass=PoolMeta):
@@ -53,27 +54,7 @@ class ConfigurationSequence(metaclass=PoolMeta):
             ('sequence_type', '=',
                 Id('purchase_requisition',
                     'sequence_type_purchase_requisition')),
-            ],
-        depends=['company'])
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-        if exist:
-            table = cls.__table_handler__(module_name)
-            exist &= table.column_exist('purchase_requisition_sequence')
-
-        super(ConfigurationSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('purchase_requisition_sequence')
-        value_names.append('purchase_requisition_sequence')
-        super(ConfigurationSequence, cls)._migrate_property(
-            field_names, value_names, fields)
+            ])
 
     @classmethod
     def default_purchase_requisition_sequence(cls):
@@ -93,52 +74,42 @@ class PurchaseRequisition(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Eval('state') != 'draft',
         }
-    _depends = ['state']
 
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
+        'company.company', "Company", required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
-            },
-        depends=['lines'] + _depends)
-    number = fields.Char('Number', readonly=True, select=True)
-    description = fields.Char(
-        'Description', states=_states, depends=_depends)
+            })
+    number = fields.Char('Number', readonly=True)
+    description = fields.Char('Description', states=_states)
     employee = fields.Many2One(
-        'company.employee', 'Employee', required=True,
-        states=_states, depends=_depends)
+        'company.employee', 'Employee', required=True, states=_states)
     supply_date = fields.Date(
         'Supply Date',
         states={
             'required': ~Eval('state').in_(['draft', 'cancelled']),
             'readonly': _states['readonly'],
-            },
-        depends=_depends)
+            })
     warehouse = fields.Many2One(
         'stock.location', 'Warehouse',
         domain=[
             ('type', '=', 'warehouse'),
             ],
-        states=_states, depends=_depends)
+        states=_states)
     currency = fields.Many2One(
         'currency.currency', 'Currency',
         states={
             'readonly': (_states['readonly']
                 | (Eval('lines', [0]) & Eval('currency'))),
-            },
-        depends=_depends)
-    currency_digits = fields.Function(
-        fields.Integer('Currency Digits'), 'on_change_with_currency_digits')
+            })
     total_amount = fields.Function(
-        fields.Numeric('Total', digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+        Monetary("Total", currency='currency', digits='currency'),
         'get_amount')
-    total_amount_cache = fields.Numeric(
-        'Total Cache', digits=(16, Eval('currency_digits', 2)),
-        depends=['currency_digits'])
+    total_amount_cache = Monetary(
+        "Total Cache", currency='currency', digits='currency')
     lines = fields.One2Many(
         'purchase.requisition.line', 'requisition', 'Lines',
-        states=_states, depends=_depends)
+        states=_states)
 
     approved_by = employee_field(
         "Approved By", states=['approved', 'processing', 'done', 'cancelled'])
@@ -152,13 +123,21 @@ class PurchaseRequisition(Workflow, ModelSQL, ModelView):
             ('processing', "Processing"),
             ('done', "Done"),
             ('cancelled', "Cancelled"),
-            ], "State", readonly=True, required=True)
+            ], "State", readonly=True, required=True, sort=False)
 
     del _states
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
         super(PurchaseRequisition, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_([
+                        'draft', 'waiting', 'approved', 'processing'])))
         cls._transitions |= set((
                 ('cancelled', 'draft'),
                 ('rejected', 'draft'),
@@ -224,6 +203,11 @@ class PurchaseRequisition(Workflow, ModelSQL, ModelView):
                 [table.state], ['cancelled'],
                 where=table.state == 'cancel'))
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @classmethod
     def default_state(cls):
         return 'draft'
@@ -248,21 +232,6 @@ class PurchaseRequisition(Workflow, ModelSQL, ModelView):
         if company:
             return Company(company).currency.id
 
-    @classmethod
-    def default_currency_digits(cls):
-        Company = Pool().get('company.company')
-        company = Transaction().context.get('company')
-        if company:
-            company = Company(company)
-            return company.currency.digits
-        return 2
-
-    @fields.depends('currency')
-    def on_change_with_currency_digits(self, name=None):
-        if self.currency:
-            return self.currency.digits
-        return 2
-
     @fields.depends('lines', 'currency')
     def on_change_with_total_amount(self):
         self.total_amount = Decimal(0)
@@ -400,11 +369,13 @@ class PurchaseRequisition(Workflow, ModelSQL, ModelView):
     def approve(cls, requisitions):
         pool = Pool()
         Configuration = pool.get('purchase.configuration')
+        transaction = Transaction()
+        context = transaction.context
         cls.store_cache(requisitions)
         config = Configuration(1)
-        with Transaction().set_context(
-                queue_name='purchase',
-                queue_scheduled_at=config.purchase_process_after):
+        with transaction.set_context(
+                queue_scheduled_at=config.purchase_process_after,
+                queue_batch=context.get('queue_batch', True)):
             cls.__queue__.process(requisitions)
 
     @classmethod
@@ -448,46 +419,43 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
     _states = {
         'readonly': Eval('purchase_requisition_state') != 'draft',
         }
-    _depends = ['purchase_requisition_state']
 
     requisition = fields.Many2One(
         'purchase.requisition', 'Requisition',
-        ondelete='CASCADE', select=True, required=True)
-    supplier = fields.Many2One(
-        'party.party', 'Supplier', states=_states, depends=_depends)
+        ondelete='CASCADE', required=True)
+    supplier = fields.Many2One('party.party', 'Supplier', states=_states)
     product = fields.Many2One(
         'product.product', 'Product',
         ondelete='RESTRICT',
         domain=[
-            ('purchasable', '=', True),
+            If((Eval('purchase_requisition_state') == 'draft')
+                & ~(Eval('quantity', 0) < 0),
+                ('purchasable', '=', True),
+                ()),
             ],
-        states=_states, depends=_depends)
+        states=_states)
     product_uom_category = fields.Function(
-        fields.Many2One('product.uom.category', "Product UOM Category"),
+        fields.Many2One(
+            'product.uom.category', "Product UoM Category",
+            help="The category of Unit of Measure for the product."),
         'on_change_with_product_uom_category')
-    description = fields.Text("Description", states=_states, depends=_depends)
+    description = fields.Text("Description", states=_states)
     summary = fields.Function(fields.Char('Summary'), 'on_change_with_summary')
     quantity = fields.Float(
-        'Quantity', digits=(16, Eval('unit_digits', 2)), required=True,
-        states=_states, depends=['unit_digits'] + _depends)
+        "Quantity", digits='unit', required=True, states=_states)
     unit = fields.Many2One(
         'product.uom', 'Unit', ondelete='RESTRICT',
         states={
             'required': Bool(Eval('product')),
             'readonly': _states['readonly'],
-            },
-        depends=['product'] + _depends)
-    unit_digits = fields.Function(
-        fields.Integer('Unit Digits'), 'on_change_with_unit_digits')
-    unit_price = fields.Numeric(
-        'Unit Price', digits=price_digits, states=_states, depends=_depends)
+            })
+    unit_price = Monetary(
+        'Unit Price', currency='currency', digits=price_digits, states=_states)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    amount = fields.Function(
-        fields.Numeric('Amount',
-            digits=(
-                16, Eval('_parent_requisition', {}).get('currency_digits', 2)),
-            depends=_depends), 'on_change_with_amount')
+    amount = fields.Function(Monetary(
+            "Amount", currency='currency', digits='currency'),
+        'on_change_with_amount')
     purchase_requests = fields.One2Many(
         'purchase.request', 'origin', 'Purchase Request', readonly=True)
     purchase_requisition_state = fields.Function(fields.Selection(
@@ -506,7 +474,6 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
                 ('category', 'in', [Eval(c) for c in unit_categories]),
                 ('category', '!=', -1)),
             ]
-        cls.unit.depends.extend(unit_categories)
 
     @classmethod
     def _unit_categories(cls):
@@ -514,13 +481,11 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
 
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
-        if self.product:
-            return self.product.default_uom_category.id
+        return self.product.default_uom_category if self.product else None
 
     @fields.depends('requisition', '_parent_requisition.currency')
     def on_change_with_currency(self, name=None):
-        if self.requisition and self.requisition.currency:
-            return self.requisition.currency.id
+        return self.requisition.currency if self.requisition else None
 
     @classmethod
     def get_purchase_requisition_states(cls):
@@ -541,14 +506,14 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
         category = self.product.purchase_uom.category
         if not self.unit or self.unit.category != category:
             self.unit = self.product.purchase_uom
-            self.unit_digits = self.product.purchase_uom.digits
 
     @fields.depends('description')
     def on_change_with_summary(self, name=None):
         return firstline(self.description or '')
 
-    @fields.depends('quantity', 'unit_price', 'unit', 'requisition',
-        '_parent_requisition.currency')
+    @fields.depends(
+        'quantity', 'unit_price',
+        'requisition', '_parent_requisition.currency')
     def on_change_with_amount(self, name=None):
         if (self.unit_price is None) or (self.quantity is None):
             return None
@@ -557,29 +522,50 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
             amount = self.requisition.currency.round(amount)
         return amount
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if self.unit:
-            return self.unit.digits
-        return 2
-
     def get_rec_name(self, name):
         pool = Pool()
         Lang = pool.get('ir.lang')
         if self.product:
             lang = Lang.get()
-            return (lang.format(
-                    '%.*f', (self.unit.digits, self.quantity or 0))
-                + '%s %s @ %s' % (
-                    self.unit.symbol, self.product.rec_name,
-                    self.requisition.rec_name))
+            return (lang.format_number_symbol(
+                    self.quantity or 0, self.unit, digits=self.unit.digits)
+                + ' %s @ %s' % (
+                    self.product.rec_name, self.requisition.rec_name))
         else:
             return self.requisition.rec_name
 
     def _get_purchase_request_product_supplier_pattern(self):
-        return {
+        pattern = {
             'company': self.requisition.company.id,
             }
+        if self.supplier:
+            pattern['party'] = self.supplier.id
+        return pattern
+
+    @property
+    def request_unit(self):
+        unit = self.unit
+        if (self.product
+                and self.product.purchase_uom.category == self.unit.category):
+            unit = self.product.purchase_uom
+        return unit
+
+    @property
+    def request_quantity(self):
+        pool = Pool()
+        Uom = pool.get('product.uom')
+        quantity = self.quantity
+        request_unit = self.request_unit
+        if (self.product
+                and request_unit
+                and request_unit.category == self.unit.category):
+            quantity = Uom.compute_qty(
+                self.unit, self.quantity, request_unit, round=True)
+        return quantity
+
+    @property
+    def request_unit_price(self):
+        return self.unit_price
 
     def compute_request(self):
         """
@@ -594,32 +580,35 @@ class PurchaseRequisitionLine(sequence_ordered(), ModelSQL, ModelView):
             return
 
         supply_date = self.requisition.supply_date
-        purchase_date = supply_date
+        supplier = None
+        purchase_date = None
 
-        if not self.supplier and self.product:
+        if self.product:
             supplier, purchase_date = Request.find_best_supplier(
                 self.product, supply_date,
                 **self._get_purchase_request_product_supplier_pattern())
-        else:
-            supplier = self.supplier
-            # TODO compute purchase_date for product_supplier
+        elif self.supplier:
+            lead_time = self.supplier.get_multivalue(
+                'supplier_lead_time', company=self.requisition.company.id)
+            if lead_time is not None:
+                purchase_date = supply_date - lead_time
 
-        uom = self.unit
-        quantity = self.quantity
+        unit = self.request_unit
+        quantity = self.request_quantity
         if (self.product
                 and self.product.purchase_uom.category == self.unit.category):
-            uom = self.product.purchase_uom
-            quantity = Uom.compute_qty(self.unit, self.quantity,
-                uom, round=True)
+            unit = self.product.purchase_uom
+            quantity = Uom.compute_qty(
+                self.unit, self.quantity, unit, round=True)
 
         return Request(
             product=self.product,
             description=self.description,
-            party=supplier,
+            party=supplier or self.supplier,
             quantity=quantity,
-            uom=uom,
+            unit=unit,
             computed_quantity=self.quantity,
-            computed_uom=self.unit,
+            computed_unit=self.unit,
             purchase_date=purchase_date,
             supply_date=supply_date,
             company=self.requisition.company,
@@ -683,8 +672,9 @@ class CreatePurchase(Wizard):
         RequisitionLine = pool.get('purchase.requisition.line')
         key = super(CreatePurchase, self)._group_purchase_line_key(request)
         if isinstance(request.origin, RequisitionLine):
-            if request.origin.unit_price:
-                key += (('unit_price', request.origin.unit_price),)
+            unit_price = request.origin.request_unit_price
+            if unit_price:
+                key += (('unit_price', unit_price),)
         return key
 
     @classmethod
